<?php



#   MultiViews
#   ==========
#   content-negotiation for nanoweb
#
#   - returns faster now for existing files
#   - .nwaccess directive: MultiViews = 1, or 0 respectively
#   - FallBackPriority = time | time desc | size | size desc | stupid
#   - removed ALL regexs in favour of string arithmetics
#   - now only accepts asterisks on the right, for example "text/*"
#     (current clients never send "*a*/x-*n*" or such things)
#   - and the Accept: "*/*" is now ignored, as it formerly lead to known
#     mime types to be favoured over unknown file types;
#     btw "*/*" never really helps in choosing a "best variant"
#   - the split_path_info() stuff is removed, as nanoweb now strips
#     it in an earlier stage (core_before_decode)
#
#   mario@erphesfurt.de



class mod_multiviews {



   function mod_multiviews() {

      $this->modtype="core_after_decode";
      $this->modname="MultiViews (content negotiation)";

      $GLOBALS["mime_enc"]["gz"] = "gzip";
      $GLOBALS["mime_enc"]["Z"] = "compress";

   }




      # fetches extensions array matching the mime regex
      # from global mime table 
      #

   function extensions($string = "text/*") {
      global $mime;

      $extensions = array();

      $compare_part = (strrpos($string, '*'));
      $string = rtrim($string, '*');

      foreach ($mime as $ext => $mimetype) {
         if ($compare_part && (strpos($mimetype, $string) === 0)
             || (!$compare_part))
         {
            $extensions[] = $ext;
         }
      }

      return($extensions);
   }




      # returns array with quality values indexed by
      # file name extensions

   function parseQualities($qstring, $q = 0.9, $divide_q = 1.7) {

      $qualities = array();

      foreach (explode(',', $qstring) as $qspart) {

         @list($mime, $quality_string) = explode(';', $qspart);

         if ($mime = trim($mime)) {

            //-- mime type or country code?
            if (strrpos($mime, '/') <= 0) {
               list($ext) = explode('_', $mime);
               $ext = array($ext);
            }
            else {
               $ext = $this->extensions($mime);
            }

            //-- extract quality value
            @list($qsarg, $new_q) = @explode('=', trim(@$quality_string));
            if ( ($qsarg[0] == "q") && !empty($new_q) ) {
               $q = trim($new_q);
            }

            //-- append fetched extensions to qualities array
            foreach ($ext as $add) {
               $qualities[$add] = $q;
            }

            $q /= $divide_q;
         }
      }

      return($qualities);
   }






   function main() {

      global $http_uri, $conf, $docroot, $htreq_headers, $mime_enc, $out_add_headers, $pri_err;
      if (access_query("reflectrewriting", 0)) { global $real_uri; }
      $enabled = strtolower(trim(access_query("multiviews", 0)));

      // quick skips
      if ( ($enabled === "0") || ($enabled == "off") ||
           file_exists($docroot . $http_uri) && substr($http_uri, strrpos($http_uri, "/"), 255) ) {
         return(1);
      }

      // determine file name to be worked on
      $file = $http_uri;
      $act_path = "";
      if ($last_slash = strrchr("/", $file)) {
         $act_path = substr($req_file, 0, $last_slash);
         $file = basename($file);
      }
      $allowed_filenames = array($file);
#DEBUG#techo("AFP == $act_path :: {$allowed_filenames[0]}");

      // add '/index.html/', because main server doesn't search variants
      if (empty($req_file)) { 
         $req_file = explode(" ", trim(implode(" ", access_query("directoryindex"))));
      }

      // find files with same basename and different extensions
      $alternative_files = array();
      $dir = opendir($docroot . $act_path);
      while($filename = readdir($dir)) {
         foreach ($allowed_filenames as $filebn) {
            $filebn .= ".";
            if (substr($filename, 0, strlen($filebn)) == $filebn) {
               $alternative_files["$filename"] = 0.0;
            }
         }
      }
      closedir($dir);

#foreach ($alternative_files as $e => $q) techo("files: '$e' => $q");
      if (empty($alternative_files)) {
         return(2);
      }  // quick skip2

      // fetch priorities
      $qualities = array_merge(
         $this->parseQualities(strtr(access_query("languagepriority", 0), " ", ", "), 0.5),
         $this->parseQualities("de, en, fr", 0.5),
         $this->parseQualities(@$htreq_headers["ACCEPT-LANGUAGE"], 1.0),
         array(
            "html" => 0.009,   // should we prefer .php instead??  ;-)
            "png"  => 0.003,
            "jpeg" => 0.002,
            "gif"  => 0.001    // ... these are micro-priorities
         ),
         $this->parseQualities(implode(", ", access_query("otherpriority")), 0.3),
         $this->parseQualities(@$htreq_headers["ACCEPT"], 1.0, 1/0.85)
      );
      foreach ($mime_enc as $ext => $encoding) {
         if (strpos("{$encoding}", $htreq_headers["ACCEPT-ENCODING"]) !== false) {
            $qualities = array_merge($qualities, array($ext => 0.5));
         }
      }

      // go thru filename extensions, and sum qualities
      foreach ($alternative_files as $filename => $file_q) {

         $file_extensions = array_slice(explode('.', $filename), 1);
         foreach ($file_extensions as $ext) {

            foreach ($qualities as $q_ext => $q) {
               if ($q_ext == $ext) {
                  $alternative_files[$filename] += $q;
               }
            }

         }
      }

      // sort
      arsort($alternative_files, SORT_NUMERIC);

      // FallBackPriority, if the first two quality values are equal
      @list($q_f0, $q_f1) = array_values($alternative_files);
      if ($q_f0 == $q_f1) {
         $n = 1;
         while ($alternative_files[$n] === $q_f0) {
            $n++;
         }
         $wf = array_splice($alternative_files, 0, $n);
         foreach ($wf as $filename => $file_q) {
            switch (trim(access_query("fallbackpriority", 0))) {
               case "newest": $wf[$filename] += filemtime($docroot . $filename) / 1000; break;
               case "oldest": $wf[$filename] += 1000 / filemtime($docroot . $filename); break;
               case "smallest": $wf[$filename] += 1000 / filesize($docroot . $filename);break;
               case "largest": $wf[$filename] += filesize($docroot . $filename) / 1000; break;
               default:
            }
         }
         arsort($wf, $SORT_NUMERIC);
         $alternative_files = array_merge($wf, $alternative_files);
      }

      // return "best" variant
      list($file) = array_keys($alternative_files);
      $real_uri=$http_uri = $act_path . $file;

      // server-driven negotiation
      $negotiation_agent = strtolower(trim(@$conf["global"]["negotiation"][0])) == "agent";
      if ( ($negotiation_agent)
           && empty($htreq_headers["NEGOTIATE"]) )
      {
        $out_add_headers["Content-Location"] = $file;
        $out_add_headers["Vary"] = "negotiate"; // which request headers affect selection
        $out_add_headers["TCN"] = "choice";

        // report encoding to client
        foreach ($mime_enc as $ext => $encoding) {
           if (strpos($file, ".{$ext}")) {
              $out_add_headers["Content-Encoding"] = $encoding;
              $GLOBALS["out_encoded"] = true;
           }
        }
      }

      // agent-driven negotiation
      else {
        techo("WARN: user agent driven content negotiation was selected");

        $out_add_headers["Alternatives"] =
        $out_add_headers["Alternates"] = implode(" , ", array_keys($alternative_files));

        $GLOBALS["pri_err"] = 300;  // HTTP 300 Choose Yourself
        $GLOBALS["out_contenttype"] = "text/html";
        $GLOBALS["lf"] = "<HTML><HEAD><TITLE>Choose Variant</TITLE></HEAD><BODY><H2>Choose Variant</H2>The document you requested exists in different variants, and your browser gives you the opportunity to select one of them:<BR><UL>";
        foreach ($alternative_files as $f => $q) {
           $GLOBALS["lf"] .= '<LI><A HREF="' . url_to_absolute($act_path . $f) . '">' . $f . '</A></LI>';
        }
        $GLOBALS["lf"] .= "</UL><HR>" . _genpage_signature() . "</BODY></HTML>";
      }

      // may help(?)
      if ($GLOBALS['path_info']) {
        $out_add_headers["Content-Base"] = $act_path;
      }

#foreach ($qualities as $e => $q) techo("qualities: '.$e' => $q");

   }


}

?>