<?php

/*

RewriteEngine
=============

tries to mimic most features of apache's mod_rewrite,
should accept standard RewriteRules from .htaccess-files

Refer to manual for desription of these directives:

  RewriteRule  RE[GE]X(.+)  REPLACEMENT/$1.html   [nocase,flags...]
  RewriteCond  test%{HTTP_ACCEPT}string   ! COND[PATT]*ERN    [nocase]

- changed to operate in a stage BEFORE all the other modules
- doesn't cache the rules from .htaccess-files
- redirection to upper directories may be buggy
- regex-backreferences in flags may not work yet
- there's a new [HEADER=X-Header:data] flag not found in apache's mod_rewrite

mario@erphesfurt.de

*/



class mod_rewrite {



   function mod_rewrite() {

      $this->modtype="core_before_decode";
      $this->modname="RewriteEngine";

   }



      # fetches rewriting regexs from .htaccess file in
      # given $path

   function read_rules($path, $rules_string = '') {

      $rules = array();
      $ht_regex = "([^\s]+|\\\\\s)";
      $accessfile = $GLOBALS['conf']['global']['accessfile'][0];

      if (is_array($rules_string) ||
          is_readable($path .= "/" . $accessfile)
          && ($rules_string = file($path)))

         foreach ($rules_string as $line) {

         #-- RewriteRULE/COND
         if ((strtoupper(substr(ltrim($line), 0, 7)) == "REWRITE") &&
             preg_match("/^\s*Rewrite(Rule|Cond)(\s+[!]\s+|\s+){$ht_regex}(\s+!\s+|\s+){$ht_regex}(\s+\[(.+)\])*/i", $line, $uu)) {

            $switch = strtoupper($uu[1]);
            $rule_regex = $uu[3];
            $rule_replacement = preg_replace('/[$](\d\d?)/', '\\\\\1', $uu[5]);
            $flags = @$uu[6];
            $rule_negate = preg_match('/!/', $uu[2] . $uu[4]) ? 1 : 0;

            // defaults
            $rule_modifiers = '';
            $rule_eval = '';
            $rule_chain = 0;

            if (! empty($flags) && preg_match('/\[(.+)\]/i', $flags, $flags)) {
               foreach( explode(",", $flags[1]) as $flag ) {
                  $flag = trim($flag);

                  if (preg_match('/^(NC|nocase|ignorecase|case)$/i', $flag)) {
                     $rule_modifiers .= 'i';
                  }
                  elseif (preg_match('/^R[edirect]*=*(\d+|temp|permanent|seeother)*/i', $flag, $uu)) {
                     $rule_eval .= ' $pri_redir = url_to_absolute("$http_uri");
                                     $this->last = 255; ';
                     switch (strtoupper(@$uu[1])) {
                        case "PERMANENT": $rq_err = "301"; break;
                        case "SEEOTHER": $rq_err = "303"; break;
                        default: if (! preg_match("/^\d+$/", @$uu[1]))
                         { $rq_err = 302; } else { $rq_err = $uu[1]; }
                     }
                     $rule_eval .= " \$rq_err = $rq_err; "; 
                  }
                  elseif (preg_match('/^F(orbidden)*/i', $flag)) {
                     $rule_eval .= ' $pri_err = 403; ';
                  }
                  elseif (preg_match('/^G(one)*/i', $flag)) {
                     $rule_eval .= ' $pri_err = 410; ';
                  }
                  elseif (preg_match('/^N(ext)*$/i', $flag)) {
                     $rule_eval .= ' $rule_no = -1; ';
                  }
                  elseif (preg_match('/^L(ast)*$/i', $flag)) {
                     $rule_eval .= ' $this->last = 255; ';
                  }
                  elseif (preg_match('/^(skip|S)=*(\d+)$/i', $flag, $uu)) {
                     $rule_eval .= ' $rule_no += {$uu[2]}; ';
                     //--stupid:--// $rule_eval .= " \$skip_rules = {$uu[2]} " .                     //    ' while( $skip_rules && $rules[++$rule_no] )                     //      if (empty($rules[$rule_no][4]))                     //         $skip_rules-- ';
                  }
                  elseif (preg_match('^C|chain|OR|ornext$', $flag)) {
                     $rule_chain = 1;
                  }
                  elseif (preg_match('/^T(ype)*=(.+)$/i', $flag, $uu)) {
                     $rule_eval .= " global \$out_contenttype;
                                     \$out_contenttype='{$uu[1]}'; ";
                  }
                  elseif (preg_match('/^(e|ENV)=([^:]+):(.+)$/i', $flag, $uu)) {
                     $rule_eval .= " putenv('{$uu[2]}={$uu[3]}'); ";
                  }
                  elseif (preg_match('/^(H|HEAD[ER]*)=([^:]+):(.+)$/i', $flag, $uu)) {
                     $rule_eval .= " \$out_add_headers['{$uu[2]}'] = '{$uu[3]}'; ";
                  }
                  elseif (preg_match('/^QSA(ppend)*$/i', $flag)) {
                     $rule_eval .= ' $QSA = @$query_string:"{$query_string}&":""; ';
                  }
                  elseif (preg_match('/^(PT|passthrough|NS|nosubreq|C|chain|P|proxy)$/i', $flag)) {
                     techo("unsupported RewriteRule-flag '$flag'");
                  }

               }
            }

            switch ($switch) {

               case "RULE":
                  $rule_regex = "\255{$rule_regex}\255$rule_modifiers";
                  $rules[] = array($rule_regex, $rule_replacement, $rule_eval, $rule_chain, $rule_negate);
                  break;

               case "COND":
                  $rule_replacement = "\255{$rule_replacement}\255$rule_modifiers";
                  $rules[] = array($rule_regex, $rule_replacement, $rule_eval, $rule_chain, $rule_negate, 1);
                  break;

            }
                         
         }

      }#--if & foreach(line)

      return($rules);

   }




      # interpolates %{ENV_VARs}, %{REGEX_FROM_LAST_COND}, and
      # some others from the $teststring

   function backreferences (&$teststring, $for_cond = 0) {

      global $htreq_headers, $vhost, $conf;

      $NANOWEB_VARS = nw_server_vars();
      
      $teststring = preg_replace('/%(\d)/ie', '$this->last_cond[\1]', $teststring);
      if ($for_cond) {
         $teststring = preg_replace('/[$](\d)/ie', '$this->last_rule[\1]', $teststring);
      }
      $teststring = preg_replace('/%{HTTP[_: ]+([-_A-Z0-9]+?)}/ie', '@$htreq_headers[strtoupper("\1")]', $teststring);
      $teststring = preg_replace('/%{ENV[_: ]+(.+?)}/ie', '@getenv("\1")', $teststring);
      $teststring = preg_replace('/%{([-_A-Z0-9]+?)}/ie', '@$NANOWEB_VARS[strtoupper("\1")]', $teststring);

   }




      # applies regexs (of the fetched .htaccess file from
      # $act_path) to the $sub_path

   function rewrite($act_path, &$sub_path) {

      global $conf, $http_uri, $docroot, $pri_redir, $rq_err, $pri_err, $add_errmsg, $query_string, $out_add_headers;
      if (@$conf["global"]["reflectrewriting"][0]) {
         global $real_uri;
      }

      // get rules for actual directory
      $rules = $this->read_rules("$docroot$act_path");

      for ($rule_no = 0; $rule_no < count($rules); $rule_no++) {

         $rule = $rules[$rule_no];
         list($rule_regex, $rule_replacement, $rule_eval,
              $rule_chain, $rule_negate, $rule_cond) = $rule;
         $rule_forced = ($rule_replacement == "-");

         // ======================================== RewriteRule ===========
         if (empty($rule_cond)) {

            // replaces %N's and $N's from last Rules/Conds
            $this->backreferences($rule_regex, 1);
            $this->backreferences($rule_replacement, 0);

            if ($rule_forced OR
                ($rule_negate XOR preg_match($rule_regex, $sub_path, $this->last_rule)) )
            {

#DEBUG#techo("preg_replace(\"$rule_regex\", \"$rule_replacement\", \"$sub_path\");");

               if (! $rule_forced) {
                  $sub_path = preg_replace($rule_regex, $rule_replacement, $sub_path);
               }
               if (! empty($rule_eval)) {
                  eval($rule_eval);
               }

               // did the rule add an querystring?
               if (preg_match('/[?]/', $sub_path)) {
                  list($sub_path, $appended_querystring) = explode('?', $sub_path);
                  $query_string = @$QSA . $appended_querystring;
               }
              
               // which nanoweb variable to put the new uri
               if ($sub_path[0] == "/") {
                  $real_uri=$http_uri = $sub_path;
                  $this->act = 0;
               }
               elseif (substr($sub_path, 0, 7) == "http://") {
                  $pri_redir = $sub_path;
                  $this->last = 255;
               }
               else {
                  $real_uri  // <- makes rewritten urls visible in errormsgs
                  = $http_uri = ($act_path ? "$act_path/" : "") . $sub_path;
               }

               $this->new_parts();   // collapse .. and . in path

            }
            elseif ($rule_chain) {     //-- on mismatch+chain skip next rules
               while (empty($rules[$rule_no][4])) {  $rule_no++; }
            }

         }

         // ========================================== RewriteCond ===========
         else {

#DEBUG#techo("condition($rule_regex, $rule_replacement)");

            // replace $N's and %N's from last Rules/Conds
            $this->backreferences($rule_regex, 1);
            $this->backreferences($rule_replacement, 1);

            $cond_matches = false || $rule_forced;
            $rule_or = 1 - $rule_chain;  // was negated for RewriteConds
            $rule_condpattern = trim(preg_replace('/[a-z]+$/', '', $rule_replacement), "\255");

            if (preg_match('/^(-[dfslFU])$/', $rule_condpattern, $uu)) {
               $filename = "$docroot$actpath/$rule_regex";
               switch (strtolower($uu[1])) {
                 case "-u":  $cond_match = ("" != implode("", $file($rule_regex)));   break;
                 case "-s":  $cond_match = (filesize($filename) > 0);   break;
                 case "-d":  $cond_match = (is_dir($filename));   break;
                 case "-l":  $cond_match = (is_link($filename));   break;
                 default:  $cond_match = (file_exists($filename));
               }
            }
            elseif (preg_match('/^([<>=])(.+)$/', $rule_condpattern, $uu)) {
               if ($uu[2] == '""')  $uu[2] = "";  // make really empty
               switch($uu[1]) {
                  case "<":  $cond_match = ($rule_regex < $uu[2]);    break;
                  case ">":  $cond_match = ($rule_regex > $uu[2]);    break;
                  default:  $cond_match = ($rule_regex == $uu[2]);
               }
            }
            else {
               $cond_match =
                  preg_match($rule_replacement, $rule_regex, $this->last_cond);
              
            }
            $cond_match = $cond_match XOR $rule_negate;
 
            // skip following RewriteConds + (first) RewriteRule(s)
            if ((! $cond_match) && (! $cond_or)) {
               while ($rules[$rule_no][5] && !$rules[$rule_no][4]) {
                  $rule_no++;   // skip RewriteCond
               }
               if (!$rules[$rule_no][5]) {
                  do {
                     $rule_no++;   // skip RewriteRules in chain
                  } while ($rules[$rule_no-1][3] && !$rules[$rule_no][5]);
               }
               $rule_no--;   // correcting, for loop also counts up this var
            }

         }

         if ($this->last)  break;

      }#--foreach(rule)
   }




   function new_parts() {

      global $http_uri;

      // canonicialize $http_uri
      $http_uri = preg_replace('#[^/]+/[.][.]#', '', $http_uri);
      $http_uri = preg_replace('#/[.]+/#', '/', $http_uri);
      $http_uri = preg_replace('#^[.]+/#', '', $http_uri);

      $http_uri = ltrim($http_uri, "/");
   }




   function main() {

      global $conf, $http_uri, $docroot;

      $this->last = false;
      $this->act = 0;
      $this->last_rule = $this->last_cond = array();

         // RewriteRules from docroot's .htaccess
         $this->rewrite($act_path = "", $http_uri);


         while (($this->act = strpos($http_uri, "/", $this->act)) !== false)
         {

            // split http_uri
            $act_path = substr($http_uri, 0, ++$this->act);
            $sub_path = substr($http_uri, $this->act, 255);


            if (is_dir($docroot . $act_path)) {

               $this->rewrite($act_path, $sub_path);

            }
            // not an directory, so we won't find any more .htaccess
            else  break 1;

            if ($this->last)  break;

         }

      return(true or false);

   }


}

?>